from enum import IntEnum

from spherov2.toy.core import Toy
from spherov2.types import ToyType


class RVR(Toy):
    toy_type = ToyType('Sphero RVR', 'RV-', 'RV', .075)

    class LEDs(IntEnum):
        RIGHT_HEADLIGHT_RED = 0
        RIGHT_HEADLIGHT_GREEN = 1
        RIGHT_HEADLIGHT_BLUE = 2
        LEFT_HEADLIGHT_RED = 3
        LEFT_HEADLIGHT_GREEN = 4
        LEFT_HEADLIGHT_BLUE = 5
        LEFT_STATUS_INDICATION_RED = 6
        LEFT_STATUS_INDICATION_GREEN = 7
        LEFT_STATUS_INDICATION_BLUE = 8
        RIGHT_STATUS_INDICATION_RED = 9
        RIGHT_STATUS_INDICATION_GREEN = 10
        RIGHT_STATUS_INDICATION_BLUE = 11
        BATTERY_DOOR_REAR_RED = 12
        BATTERY_DOOR_REAR_GREEN = 13
        BATTERY_DOOR_REAR_BLUE = 14
        BATTERY_DOOR_FRONT_RED = 15
        BATTERY_DOOR_FRONT_GREEN = 16
        BATTERY_DOOR_FRONT_BLUE = 17
        POWER_BUTTON_FRONT_RED = 18
        POWER_BUTTON_FRONT_GREEN = 19
        POWER_BUTTON_FRONT_BLUE = 20
        POWER_BUTTON_REAR_RED = 21
        POWER_BUTTON_REAR_GREEN = 22
        POWER_BUTTON_REAR_BLUE = 23
        LEFT_BRAKELIGHT_RED = 24
        LEFT_BRAKELIGHT_GREEN = 25
        LEFT_BRAKELIGHT_BLUE = 26
        RIGHT_BRAKELIGHT_RED = 27
        RIGHT_BRAKELIGHT_GREEN = 28
        RIGHT_BRAKELIGHT_BLUE = 29
        UNDERCARRIAGE_WHITE = 30



import importlib
from functools import partial
from typing import List, Type

from spherov2.toy.core import Toy
from spherov2.toy.r2d2 import R2D2
from spherov2.toy.r2q5 import R2Q5


class ToyNotFoundError(Exception):
    ...


def all_toys(cls=Toy):
    subtypes = cls.__subclasses__()
    yield cls
    for sub in subtypes:
        yield from all_toys(sub)


def find_toys(timeout=5.0, toy_types: List[Type[Toy]] = None, toy_names: List[str] = None, adapter=None):
    if adapter is None:
        adapter = importlib.import_module('spherov2.adapter.bleak_adapter').BleakAdaptor
    toys = adapter.scan_toys(timeout)
    if toy_types is None:
        toy_types = list(all_toys())
    ret = []
    if toy_names is not None:
        toy_names = set(toy_names)
    for toy in toys:
        if toy_names is not None and toy.name not in toy_names:
            continue
        for toy_cls in toy_types:
            toy_type = toy_cls.toy_type
            if toy.name.startswith(toy_type.filter_prefix) and \
                    (toy_type.prefix is None or toy.name.startswith(toy_type.prefix)):
                ret.append(toy_cls(toy, adapter))
                break
    return ret


def find_toy(*args, toy_name: str = None, **kwargs):
    toys = find_toys(*args, toy_names=[toy_name] if toy_name else None, **kwargs)
    if not toys:
        raise ToyNotFoundError
    return toys[0]


find_R2D2 = partial(find_toy, toy_types=[R2D2])
find_R2Q5 = partial(find_toy, toy_types=[R2Q5])



from functools import partial

from spherov2.packet import Packet


class SystemInfo:
    __encode = partial(Packet, device_id=17)

    @staticmethod
    def get_main_app_version(target_id=None):
        return SystemInfo.__encode(command_id=0, target_id=target_id)

    @staticmethod
    def get_mac_address(target_id=None):
        return SystemInfo.__encode(command_id=6, target_id=target_id)

    @staticmethod
    def get_stats_id(target_id=None):
        return SystemInfo.__encode(command_id=19, target_id=target_id)

    @staticmethod
    def get_secondary_main_app_version(target_id=None):
        return SystemInfo.__encode(command_id=23, target_id=target_id)

    secondary_main_app_version_notify = (17, 24, 0xff)

    @staticmethod
    def get_three_character_sku(target_id=None):
        return SystemInfo.__encode(command_id=40, target_id=target_id)



import struct
from enum import IntEnum
from functools import partial

from spherov2.helper import to_bytes
from spherov2.packet import Packet


class R2DoLegActions(IntEnum):
    UNKNOWN = 0
    THREE_LEGS = 1
    TWO_LEGS = 2
    WADDLE = 3
    TRANSITIONING = 4


class R2LegActions(IntEnum):
    STOP = 0
    THREE_LEGS = 1
    TWO_LEGS = 2
    WADDLE = 3


class Animatronic:
    __encode = partial(Packet, device_id=23)

    @staticmethod
    def play_animation(animation: IntEnum, target_id=None):
        return Animatronic.__encode(command_id=5, data=to_bytes(animation, 2), target_id=target_id)

    @staticmethod
    def perform_leg_action(leg_action: R2LegActions, target_id=None):
        return Animatronic.__encode(command_id=13, data=[leg_action], target_id=target_id)

    @staticmethod
    def set_head_position(head_position: float, target_id=None):
        return Animatronic.__encode(command_id=15, data=struct.pack('>f', head_position), target_id=target_id)

    play_animation_complete_notify = (23, 17, 0xff)

    @staticmethod
    def set_leg_position(leg_position: float, target_id=None):
        return Animatronic.__encode(command_id=21, data=struct.pack('>f', leg_position), target_id=target_id)

    @staticmethod
    def get_leg_position(target_id=None):
        return Animatronic.__encode(command_id=22, target_id=target_id)

    leg_action_complete_notify = (23, 38, 0xff)

    @staticmethod
    def enable_leg_action_notify(enable: bool, target_id=None):
        return Animatronic.__encode(command_id=42, data=[int(enable)], target_id=target_id)

    @staticmethod
    def stop_animation(target_id=None):
        return Animatronic.__encode(command_id=43, target_id=target_id)



import struct
from collections import OrderedDict
from typing import Callable, List

from spherov2.commands.async_ import Async
from spherov2.commands.core import Core, PowerStates, IntervalOptions
from spherov2.commands.sphero import Sphero as SpheroCmd, CollisionDetectionMethods, RollModes, ReverseFlags, \
    RawMotorModes
from spherov2.listeners.core import VersionsArgs, BluetoothInfoArgs, PowerStateArgs
from spherov2.listeners.sensor import CollisionDetectedArgs
from spherov2.toy.core import Toy, ToySensor
from spherov2.types import ToyType


class Sphero(Toy):
    toy_type = ToyType('SPRK/2.0', None, 'Sphero', .06)

    sensors = OrderedDict(
        attitude=OrderedDict(
            pitch=ToySensor(0x40000, -179., 180.),
            roll=ToySensor(0x20000, -179., 180.),
            yaw=ToySensor(0x10000, -179., 180.)
        ),
        accelerometer=OrderedDict(
            x=ToySensor(0x8000, -32768., 32767., lambda x: x / 4096),
            y=ToySensor(0x4000, -32768., 32767., lambda x: x / 4096),
            z=ToySensor(0x2000, -32768., 32767., lambda x: x / 4096)
        ),
        gyroscope=OrderedDict(
            x=ToySensor(0x1000, -20000., 20000., lambda x: x * .1),
            y=ToySensor(0x800, -20000., 20000., lambda x: x * .1),
            z=ToySensor(0x400, -20000., 20000., lambda x: x * .1)
        ),
        back_emf=OrderedDict(
            left=ToySensor(0x40, -32768., 32767.),
            right=ToySensor(0x20, -32768., 32767.),
        )
    )

    extended_sensors = OrderedDict(
        quaternion=OrderedDict(
            x=ToySensor(0x80000000, -10000., 10000., lambda x: x / 10000),
            y=ToySensor(0x40000000, -10000., 10000., lambda x: x / 10000),
            z=ToySensor(0x20000000, -10000., 10000., lambda x: x / 10000),
            w=ToySensor(0x10000000, -10000., 10000., lambda x: x / 10000)
        ),
        locator=OrderedDict(
            x=ToySensor(0x8000000, -32768., 32767.),
            y=ToySensor(0x4000000, -32768., 32767.),
        ),
        accel_one=OrderedDict(accel_one=ToySensor(0x2000000, 0., 8000.)),
        velocity=OrderedDict(
            x=ToySensor(0x1000000, -32768., 32767., lambda x: x * .1),
            y=ToySensor(0x800000, -32768., 32767., lambda x: x * .1),
        ),
        speed=OrderedDict(speed=ToySensor(0x400000, 0., 32767.)),
    )

    def ping(self):
        self._execute(Core.ping())

    def get_versions(self) -> VersionsArgs:
        unpacked = struct.unpack('>8B', bytearray(self._execute(Core.get_versions()).data))
        return VersionsArgs(
            record_version=unpacked[0], model_number=unpacked[1], hardware_version_code=unpacked[2],
            main_app_version_major=unpacked[3], main_app_version_minor=unpacked[4],
            bootloader_version='%d.%d' % (unpacked[5] >> 4, unpacked[5] & 0xf),
            orb_basic_version='%d.%d' % (unpacked[6] >> 4, unpacked[6] & 0xf),
            overlay_version='%d.%d' % (unpacked[7] >> 4, unpacked[7] & 0xf),
        )

    def set_bluetooth_name(self, name: str):
        self._execute(Core.set_bluetooth_name(name))

    def get_bluetooth_info(self) -> BluetoothInfoArgs:
        data = self._execute(Core.get_bluetooth_info()).data
        name = ''
        i = 0
        while i < len(data) and data[i] != 0:
            name += chr(data[i])
            i += 1
        while i < len(data) and data[i] == 0:
            i += 1
        address = ''
        while i < len(data) and data[i] != 0:
            address += chr(data[i])
            i += 1
        return BluetoothInfoArgs(name=name, address=address)

    def get_power_state(self) -> PowerStateArgs:
        unpacked = struct.unpack('>2B3H', bytearray(self._execute(Core.get_power_state()).data))
        return PowerStateArgs(record_version=unpacked[0], state=PowerStates(unpacked[1]), voltage=unpacked[2] / 100,
                              number_of_charges=unpacked[3], time_since_last_charge=unpacked[4])

    def enable_battery_state_changed_notify(self, enable: bool):
        self._execute(Core.enable_battery_state_changed_notify(enable))

    def sleep(self, interval_option: IntervalOptions, unk: int, unk2: int):  # FIXME: unknown arg name
        self._execute(Core.sleep(interval_option, unk, unk2))

    def set_inactivity_timeout(self, timeout: int):
        self._execute(Core.set_inactivity_timeout(timeout))

    # TODO: getChargerState, jumpToBootloader, beginReflash, hereIsPage, jumpToMain

    def set_heading(self, heading: int):
        self._execute(SpheroCmd.set_heading(heading))

    def set_stabilization(self, stabilize: bool):
        self._execute(SpheroCmd.set_stabilization(stabilize))

    # TODO: setRotationRate, getChassisId, selfLevel

    def set_data_streaming(self, interval, num_samples_per_packet, mask, count, extended_mask):
        self._execute(SpheroCmd.set_data_streaming(interval, num_samples_per_packet, mask, count, extended_mask))

    def configure_collision_detection(self, collision_detection_method: CollisionDetectionMethods,
                                      x_threshold, y_threshold, x_speed, y_speed, dead_time):
        self._execute(SpheroCmd.configure_collision_detection(
            collision_detection_method, x_threshold, y_threshold, x_speed, y_speed, dead_time
        ))

    def configure_locator(self, flags, x, y, yaw_tare):  # name from web
        self._execute(SpheroCmd.configure_locator(flags, x, y, yaw_tare))

    # TODO: getTemperature

    def set_main_led(self, r, g, b):
        self._execute(SpheroCmd.set_main_led(r, g, b))

    def set_back_led_brightness(self, brightness):
        self._execute(SpheroCmd.set_back_led_brightness(brightness))

    def roll(self, speed, heading, roll_mode: RollModes, reverse_flag: ReverseFlags):
        self._execute(SpheroCmd.roll(speed, heading, roll_mode, reverse_flag))

    # TODO: boost

    def set_raw_motors(self, left_mode: RawMotorModes, left_speed, right_mode: RawMotorModes, right_speed):
        self._execute(SpheroCmd.set_raw_motors(left_mode, left_speed, right_mode, right_speed))

    # TODO: setMotionTimeout, setPersistentOptions, getPersistentOptions, setTemporaryOptions, getTemporaryOptions

    def add_battery_state_changed_notify_listener(self, listener: Callable[[PowerStates], None]):
        self._add_listener(Async.battery_state_changed_notify, lambda p: listener(PowerStates(p.data[0])))

    def remove_battery_state_changed_notify_listener(self, listener):
        self._remove_listener(Async.battery_state_changed_notify, listener)

    def add_sensor_streaming_data_notify_listener(self, listener: Callable[[List[int]], None]):
        self._add_listener(Async.sensor_streaming_data_notify,
                           lambda p: listener(list(struct.unpack('>%dH' % (len(p.data) // 2), bytearray(p.data)))))

    def remove_sensor_streaming_data_notify_listener(self, listener):
        self._remove_listener(Async.sensor_streaming_data_notify, listener)

    def add_will_sleep_notify_listener(self, listener: Callable[[], None]):
        self._add_listener(Async.will_sleep_notify, lambda _: listener())

    def remove_will_sleep_notify_listener(self, listener):
        self._remove_listener(Async.will_sleep_notify, listener)

    def add_collision_detected_notify_listener(self, listener: Callable[[CollisionDetectedArgs], None]):
        def __process(packet):
            unpacked = struct.unpack('>3HB3HBL', bytearray(packet.data))
            listener(CollisionDetectedArgs(acceleration_x=unpacked[0] / 4096, acceleration_y=unpacked[1] / 4096,
                                           acceleration_z=unpacked[2] / 4096, x_axis=bool(unpacked[3] & 1),
                                           y_axis=bool(unpacked[3] & 2), power_x=unpacked[4], power_y=unpacked[5],
                                           power_z=unpacked[6], speed=unpacked[7], time=unpacked[8] / 1000))

        self._add_listener(Async.collision_detected_notify, __process)

    def remove_collision_detected_notify_listener(self, listener):
        self._remove_listener(Async.collision_detected_notify, listener)

    def add_gyro_max_notify_listener(self, listener: Callable[[int], None]):
        self._add_listener(Async.gyro_max_notify, lambda p: listener(p.data[0]))

    def remove_gyro_max_notify_listener(self, listener):
        self._remove_listener(Async.gyro_max_notify, listener)

    def add_did_sleep_notify_listener(self, listener: Callable[[], None]):
        self._add_listener(Async.gyro_max_notify, lambda p: listener())

    def remove_did_sleep_notify_listener(self, listener):
        self._remove_listener(Async.gyro_max_notify, listener)



class Async:
    battery_state_changed_notify = (0xff, 1, 0xff)
    sensor_streaming_data_notify = (0xff, 3, 0xff)
    will_sleep_notify = (0xff, 5, 0xff)
    collision_detected_notify = (0xff, 7, 0xff)
    gyro_max_notify = (0xff, 12, 0xff)
    did_sleep_notify = (0xff, 20, 0xff)



from enum import IntFlag
from functools import partial

from spherov2.packet import Packet


class PendingUpdateFlags(IntFlag):
    nordic = 0b1
    st = 0b10
    esp2886 = 0b100
    audio = 0b1000
    animations = 0b10000
    st_bootloader = 0b100000


class Firmware:
    __encode = partial(Packet, device_id=29)

    @staticmethod
    def get_pending_update_flags(target_id=None):
        return Firmware.__encode(command_id=13, target_id=target_id)



from typing import NamedTuple

from spherov2.commands.core import PowerStates


class VersionsArgs(NamedTuple):
    record_version: int
    model_number: int
    hardware_version_code: int
    main_app_version_major: int
    main_app_version_minor: int
    bootloader_version: str
    orb_basic_version: str
    overlay_version: str


class BluetoothInfoArgs(NamedTuple):
    name: str
    address: str


class PowerStateArgs(NamedTuple):
    record_version: int
    state: PowerStates
    voltage: float
    number_of_charges: int
    time_since_last_charge: int



from enum import IntFlag, IntEnum
from functools import partial

from spherov2.helper import to_bytes
from spherov2.packet import Packet


class DriveFlags(IntFlag):
    FORWARD = 0b0
    BACKWARD = 0b1
    TURBO = 0b10


class StabilizationIndexes(IntEnum):
    NO_CONTROL_SYSTEM = 0
    FULL_CONTROL_SYSTEM = 1
    PITCH_CONTROL_SYSTEM = 2
    ROLL_CONTROL_SYSTEM = 3
    YAW_CONTROL_SYSTEM = 4
    SPEED_AND_YAW_CONTROL_SYSTEM = 5


class RawMotorModes(IntFlag):
    OFF = 0b0
    FORWARD = 0b1
    REVERSE = 0b10


class Driving:
    __encode = partial(Packet, device_id=22)

    @staticmethod
    def set_raw_motors(left_mode: RawMotorModes, left_speed, right_mode: RawMotorModes, right_speed, target_id=None):
        return Driving.__encode(command_id=1, data=[left_mode, left_speed, right_mode, right_speed],
                                target_id=target_id)

    @staticmethod
    def reset_yaw(target_id=None):
        return Driving.__encode(command_id=6, target_id=target_id)

    @staticmethod
    def drive_with_heading(speed, heading, drive_flags: DriveFlags, target_id=None):
        return Driving.__encode(command_id=7, data=[speed, *to_bytes(heading, 2), drive_flags], target_id=target_id)

    @staticmethod
    def set_stabilization(stabilization_index: StabilizationIndexes, target_id=None):
        return Driving.__encode(command_id=12, data=[stabilization_index], target_id=target_id)



from enum import IntFlag
from functools import partial

from spherov2.helper import to_bytes
from spherov2.packet import Packet


class AudioPlaybackModes(IntFlag):
    PLAY_IMMEDIATELY = 0b0
    PLAY_ONLY_IF_NOT_PLAYING = 0b1
    PLAY_AFTER_CURRENT_SOUND = 0b10


class IO:
    __encode = partial(Packet, device_id=26)

    @staticmethod
    def play_audio_file(sound, playback_mode: AudioPlaybackModes, target_id=None):
        return IO.__encode(command_id=7, data=[*to_bytes(sound, 2), playback_mode], target_id=target_id)

    @staticmethod
    def set_audio_volume(volume, target_id=None):
        return IO.__encode(command_id=8, data=[volume], target_id=target_id)

    @staticmethod
    def stop_all_audio(target_id=None):
        return IO.__encode(command_id=10, target_id=target_id)

    @staticmethod
    def set_all_leds_with_16_bit_mask(mask, values, target_id=None):
        return IO.__encode(command_id=14, data=[*to_bytes(mask, 2), *values], target_id=target_id)

    @staticmethod
    def start_idle_led_animation(target_id=None):
        return IO.__encode(command_id=25, target_id=target_id)



import threading
import time
from collections import defaultdict, OrderedDict
from concurrent import futures
from dataclasses import dataclass
from queue import SimpleQueue
from typing import Callable

from spherov2.packet import Packet, Collector
from spherov2.toy.consts import CharacteristicUUID
from spherov2.types import ToyType


class CommandExecuteError(Exception):
    ...


@dataclass
class ToySensor:
    bit: int
    min_value: float
    max_value: float
    modifier: Callable[[float], float] = None


class Toy:
    toy_type = ToyType('Robot', None, 'Sphero', .06)
    sensors = OrderedDict()
    extended_sensors = OrderedDict()

    def __init__(self, toy, adapter_cls):
        self.address = toy.address
        self.name = toy.name

        self.__adapter = None
        self.__adapter_cls = adapter_cls
        self.__decoder = Collector(self.__new_packet)
        self.__waiting = defaultdict(SimpleQueue)
        self.__listeners = defaultdict(set)

        self.__thread = None
        self.__packet_queue = SimpleQueue()

    def __enter__(self):
        if self.__adapter is not None:
            raise RuntimeError('Toy already in context manager')
        self.__adapter = self.__adapter_cls(self.address)
        self.__thread = threading.Thread(target=self.__process_packet)
        try:
            self.__adapter.set_callback(CharacteristicUUID.api_v2.value, self.__api_read)
            self.__adapter.write(CharacteristicUUID.anti_dos.value, b'usetheforce...band')
            self.__thread.start()
        except:
            self.__exit__(None, None, None)
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__adapter.close()
        self.__adapter = None
        if self.__thread.is_alive():
            self.__packet_queue.put(None)
            self.__thread.join()
        self.__packet_queue = SimpleQueue()

    def __process_packet(self):
        while self.__adapter is not None:
            payload = self.__packet_queue.get()
            if payload is None:
                break
            # print('request ' + ' '.join([hex(c) for c in payload]))
            self.__adapter.write(CharacteristicUUID.api_v2.value, payload)
            time.sleep(self.toy_type.cmd_safe_interval)

    def _execute(self, packet: Packet) -> Packet:
        if self.__adapter is None:
            raise RuntimeError('Use toys in context manager')
        self.__packet_queue.put(packet.build())
        return self._wait_packet(packet.id)

    def _wait_packet(self, key, timeout=10.0):
        future = futures.Future()
        self.__waiting[key].put(future)
        packet = future.result(timeout)
        if packet.error != Packet.Error.success:
            raise CommandExecuteError(packet.error)
        return packet

    def _add_listener(self, key, notifier: Callable[[Packet], None]):
        self.__listeners[key].add(notifier)

    def _remove_listener(self, key, notifier: Callable[[Packet], None]):
        self.__listeners[key].remove(notifier)

    def __api_read(self, char, data):
        self.__decoder.add(data)

    def __new_packet(self, packet: Packet):
        # print('response ' + ' '.join([hex(c) for c in packet.build()]))
        key = packet.id
        queue = self.__waiting[key]
        while not queue.empty():
            queue.get().set_result(packet)
        for f in self.__listeners[key]:
            threading.Thread(target=f, args=(packet,)).start()



from enum import IntEnum
from functools import partial

from spherov2.helper import to_bytes
from spherov2.packet import Packet


class PowerStates(IntEnum):
    UNKNOWN = 0
    CHARGING = 1
    OK = 2
    LOW = 3
    CRITICAL = 4


class IntervalOptions(IntEnum):
    NONE = 0
    DEEP_SLEEP = 0xffff


class Core:
    __encode = partial(Packet, device_id=0)

    @staticmethod
    def ping(target_id=None):
        return Core.__encode(command_id=1, target_id=target_id)

    @staticmethod
    def get_versions(target_id=None):
        return Core.__encode(command_id=2, target_id=target_id)

    @staticmethod
    def set_bluetooth_name(name: str, target_id=None):
        return Core.__encode(command_id=16, data=name.encode('ascii') + bytes(1), target_id=target_id)

    @staticmethod
    def get_bluetooth_info(target_id=None):
        return Core.__encode(command_id=17, target_id=target_id)

    @staticmethod
    def get_power_state(target_id=None):
        return Core.__encode(command_id=32, target_id=target_id)

    @staticmethod
    def enable_battery_state_changed_notify(enable: bool, target_id=None):
        return Core.__encode(command_id=33, data=[int(enable)], target_id=target_id)

    @staticmethod
    def sleep(interval_option: IntervalOptions, unk: int, unk2: int, target_id=None):
        return Core.__encode(command_id=34, data=[interval_option, unk, *to_bytes(unk2, 2)], target_id=target_id)

    @staticmethod
    def set_inactivity_timeout(timeout: int, target_id=None):
        return Core.__encode(command_id=37, data=to_bytes(timeout, 2), target_id=target_id)



from enum import IntEnum
from functools import partial

from spherov2.helper import to_bytes
from spherov2.packet import Packet


class CollisionDetectionMethods(IntEnum):
    OFF = 0
    DEFAULT = 1


class RollModes(IntEnum):
    STOP = 0
    GO = 1
    CALIBRATE = 2


class ReverseFlags(IntEnum):
    OFF = 0
    ON = 1


class RawMotorModes(IntEnum):
    OFF = 0
    FORWARD = 1
    REVERSE = 2
    BRAKE = 3
    IGNORE = 4


class Sphero:
    __encode = partial(Packet, device_id=2)

    @staticmethod
    def set_heading(heading: int, target_id=None):
        return Sphero.__encode(command_id=1, data=to_bytes(heading, 2), target_id=target_id)

    @staticmethod
    def set_stabilization(stabilize: bool, target_id=None):
        return Sphero.__encode(command_id=2, data=[int(stabilize)], target_id=target_id)

    @staticmethod
    def set_data_streaming(interval, num_samples_per_packet, mask, count, extended_mask, target_id=None):
        return Sphero.__encode(command_id=17,
                               data=[*to_bytes(interval, 2), *to_bytes(num_samples_per_packet, 2),
                                     *to_bytes(mask, 4), count & 0xff, *to_bytes(extended_mask, 4)],
                               target_id=target_id)

    @staticmethod
    def configure_collision_detection(collision_detection_method: CollisionDetectionMethods,
                                      x_threshold, y_threshold, x_speed, y_speed, dead_time, target_id=None):
        return Sphero.__encode(command_id=18,
                               data=[collision_detection_method, x_threshold, y_threshold, x_speed, y_speed, dead_time],
                               target_id=target_id)

    @staticmethod
    def configure_locator(flags, x, y, yaw_tare, target_id=None):
        return Sphero.__encode(command_id=19, data=[flags, *to_bytes(x, 2), *to_bytes(y, 2), *to_bytes(yaw_tare, 2)],
                               target_id=target_id)

    @staticmethod
    def set_main_led(r, g, b, target_id=None):
        return Sphero.__encode(command_id=32, data=[r, g, b], target_id=target_id)

    @staticmethod
    def set_back_led_brightness(brightness, target_id=None):
        return Sphero.__encode(command_id=33, data=[brightness], target_id=target_id)

    @staticmethod
    def roll(speed, heading, roll_mode: RollModes, reverse_flag: ReverseFlags, target_id=None):
        return Sphero.__encode(command_id=48, data=[speed, *to_bytes(heading, 2), roll_mode, reverse_flag],
                               target_id=target_id)

    @staticmethod
    def set_raw_motors(left_mode: RawMotorModes, left_speed, right_mode: RawMotorModes, right_speed, target_id=None):
        return Sphero.__encode(command_id=51, data=[left_mode, left_speed, right_mode, right_speed],
                               target_id=target_id)



import threading
from enum import IntEnum
from typing import Dict, List, Callable

from spherov2.commands.driving import DriveFlags
from spherov2.commands.driving import RawMotorModes as DriveRawMotorModes
from spherov2.controls.enums import RawMotorModes
from spherov2.toy.core import Toy


class DriveControl:
    def __init__(self, toy: Toy):
        self.__toy = toy
        self.__is_boosting = False

    def roll_start(self, heading, speed):
        flag = DriveFlags.FORWARD
        if speed < 0:
            flag = DriveFlags.BACKWARD
            heading += 180
        if self.__is_boosting:
            flag |= DriveFlags.TURBO
        speed = min(255, abs(speed))
        heading %= 360
        self.__toy.drive_with_heading(speed, heading, flag)

    def roll_stop(self, heading):
        self.roll_start(heading, 0)

    def set_stabilization(self, stabilize):
        self.__toy.set_stabilization(stabilize)

    def set_raw_motors(self, left_mode, left_speed, right_mode, right_speed):
        if left_mode == RawMotorModes.FORWARD:
            left_drive_mode = DriveRawMotorModes.FORWARD
        elif left_mode == RawMotorModes.REVERSE:
            left_drive_mode = DriveRawMotorModes.REVERSE
        else:
            left_drive_mode = DriveRawMotorModes.OFF

        if right_mode == RawMotorModes.FORWARD:
            right_drive_mode = DriveRawMotorModes.FORWARD
        elif right_mode == RawMotorModes.REVERSE:
            right_drive_mode = DriveRawMotorModes.REVERSE
        else:
            right_drive_mode = DriveRawMotorModes.OFF

        self.__toy.set_raw_motors(left_drive_mode, left_speed, right_drive_mode, right_speed)

    def reset_header(self):
        self.__toy.reset_yaw()


class LedControl:
    def __init__(self, toy: Toy):
        self.__toy = toy

    def set_leds(self, mapping: Dict[IntEnum, int]):
        mask = 0
        led_values = []
        for e in self.__toy.LEDs:
            if e in mapping:
                mask |= 1 << e
                led_values.append(mapping[e])
        if mask:
            if hasattr(self.__toy, 'set_all_leds_with_32_bit_mask'):
                self.__toy.set_all_leds_with_32_bit_mask(mask, led_values)
            elif hasattr(self.__toy, 'set_all_leds_with_16_bit_mask'):
                self.__toy.set_all_leds_with_16_bit_mask(mask, led_values)
            elif hasattr(self.__toy, 'set_all_leds_with_8_bit_mask'):
                self.__toy.set_all_leds_with_8_bit_mask(mask, led_values)


class SensorControl:
    def __init__(self, toy: Toy):
        self.__toy = toy
        toy.add_sensor_streaming_data_notify_listener(self.__process_sensor_stream_data)

        self.__count = 0
        self.__interval = 250
        self.__enabled = {}
        self.__enabled_extended = {}
        self.__listeners = set()

    def __process_sensor_stream_data(self, sensor_data: List[float]):
        data = {}

        def __new_data():
            n = {}
            for name, component in components.items():
                d = sensor_data.pop(0)
                if component.modifier:
                    d = component.modifier(d)
                n[name] = d
            data[sensor] = n

        for sensor, components in self.__toy.sensors.items():
            if sensor in self.__enabled:
                __new_data()
        for sensor, components in self.__toy.extended_sensors.items():
            if sensor in self.__enabled_extended:
                __new_data()

        for f in self.__listeners:
            threading.Thread(target=f, args=(data,)).start()

    def add_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.add(listener)

    def remove_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.remove(listener)

    def set_count(self, count: int):
        if count > 0:
            self.__count = count
            self.update()

    def set_interval(self, interval: int):
        if interval >= 0:
            self.__interval = interval
            self.update()

    def update(self):
        sensors_mask = extended_sensors_mask = 0
        for sensor in self.__enabled.values():
            for component in sensor.values():
                sensors_mask |= component.bit
        for sensor in self.__enabled_extended.values():
            for component in sensor.values():
                extended_sensors_mask |= component.bit
        self.__toy.set_sensor_streaming_mask(0, self.__count, sensors_mask)
        self.__toy.set_extended_sensor_streaming_mask(extended_sensors_mask)
        self.__toy.set_sensor_streaming_mask(self.__interval, self.__count, sensors_mask)

    def enable(self, *sensors):
        for sensor in sensors:
            if sensor in self.__toy.sensors:
                self.__enabled[sensor] = self.__toy.sensors[sensor]
            elif sensor in self.__toy.extended_sensors:
                self.__enabled_extended[sensor] = self.__toy.extended_sensors[sensor]
        self.update()

    def disable(self, *sensors):
        for sensor in sensors:
            self.__enabled.pop(sensor, None)
            self.__enabled_extended.pop(sensor, None)
        self.update()

    def disable_all(self):
        self.__enabled.clear()
        self.__enabled_extended.clear()
        self.update()



from functools import partial

from spherov2.packet import Packet


class APIAndShell:
    __encode = partial(Packet, device_id=16)

    @staticmethod
    def ping(data, target_id=None):
        return APIAndShell.__encode(command_id=0, data=data, target_id=target_id)



from enum import IntFlag, IntEnum
from typing import List, Tuple


class PacketDecodingException(Exception):
    ...


class Packet:
    """
    Packet structure, from https://sdk.sphero.com/docs/api_spec/general_api
    ---------------------------------
    - start      [1 byte]
    - flags      [1 byte]
    - target_id  [1 byte] (optional)
    - source_id  [1 byte] (optional)
    - device_id  [1 byte]
    - command_id [1 byte]
    - sequence   [1 byte]
    - error      [1 byte] (at response)
    - data       [n byte] (zero or more)
    - checksum   [1 byte]
    - end        [1 byte]
    ---------------------------------
    """

    class Flags(IntFlag):
        is_response = 0b1
        requests_response = 0b10
        requests_only_error_response = 0b100
        is_activity = 0b1000
        has_target_id = 0b10000
        has_source_id = 0b100000
        unused = 0b1000000
        extended_flags = 0b10000000

    class Encoding(IntEnum):
        escape = 0xAB
        start = 0x8D
        end = 0xD8
        escaped_escape = 0x23
        escaped_start = 0x05
        escaped_end = 0x50

    class Error(IntEnum):
        success = 0x00
        bad_device_id = 0x01
        bad_command_id = 0x02
        not_yet_implemented = 0x03
        command_is_restricted = 0x04
        bad_data_length = 0x05
        command_failed = 0x06
        bad_parameter_value = 0x07
        busy = 0x08
        bad_target_id = 0x09
        target_unavailable = 0x0a

    __sequence = -1

    def __init__(self, device_id, command_id,
                 target_id=None, source_id=None, data=None, flags=None, error=None, sequence=None):
        self.flags = flags
        if flags is None:
            self.flags = Packet.Flags.requests_response | Packet.Flags.is_activity
            if target_id is not None:
                self.flags |= Packet.Flags.has_source_id | Packet.Flags.has_target_id
                source_id = 0x1

        self.target_id = target_id
        self.source_id = source_id
        self.device_id = device_id
        self.command_id = command_id
        self.sequence = self.generate_sequence() if sequence is None else sequence
        self.data = data or []
        self.error = error

    @classmethod
    def generate_sequence(cls):
        cls.__sequence = (cls.__sequence + 1) % 0xff
        return cls.__sequence

    @property
    def id(self) -> Tuple:
        return self.device_id, self.command_id, self.sequence

    @staticmethod
    def __unescape_response_data(response_data) -> List[int]:
        raw_data = []

        iter_response_data = iter(response_data)
        for b in iter_response_data:
            if b == Packet.Encoding.escape:
                b = next(iter_response_data, None)
                if b == Packet.Encoding.escaped_escape:
                    b = Packet.Encoding.escape
                elif b == Packet.Encoding.escaped_start:
                    b = Packet.Encoding.start
                elif b == Packet.Encoding.escaped_end:
                    b = Packet.Encoding.end
                else:
                    raise PacketDecodingException('Unexpected escaping byte')

            raw_data.append(b)

        return raw_data

    @staticmethod
    def from_response(response_data) -> 'Packet':
        response_data = Packet.__unescape_response_data(response_data)
        start, flags, *data, checksum, end = response_data
        if start != Packet.Encoding.start:
            raise PacketDecodingException('Unexpected start of packet')
        if end != Packet.Encoding.end:
            raise PacketDecodingException('Unexpected end of packet')

        target_id = None
        if flags & Packet.Flags.has_target_id:
            target_id = data.pop(0)

        source_id = None
        if flags & Packet.Flags.has_source_id:
            source_id = data.pop(0)

        device_id, command_id, sequence, *data = data

        error = Packet.Error.success
        if flags & Packet.Flags.is_response:
            error = Packet.Error(data.pop(0))

        packet = Packet(
            flags=flags,
            target_id=target_id,
            source_id=source_id,
            device_id=device_id,
            command_id=command_id,
            sequence=sequence,
            error=error,
            data=data,
        )

        calc_checksum = packet.checksum
        if calc_checksum != checksum:
            raise PacketDecodingException(
                f'Bad response checksum. (Expected: {checksum:#04x}, obtained: {calc_checksum:#04x})')

        return packet

    @property
    def packet_payload(self) -> List[int]:
        ret = [self.flags]

        if self.flags & Packet.Flags.has_target_id:
            ret.append(self.target_id)

        if self.flags & Packet.Flags.has_source_id:
            ret.append(self.source_id)

        ret.extend((self.device_id, self.command_id, self.sequence))

        if self.flags & Packet.Flags.is_response:
            ret.append(self.error)

        ret.extend(self.data)
        return ret

    @property
    def checksum(self) -> int:
        return 0xff - (sum(self.packet_payload) & 0xff)

    def build(self) -> List[int]:
        full_packet = [
            *self.packet_payload,
            self.checksum,
        ]

        escaped_full_packet = []
        for i in full_packet:
            if i == Packet.Encoding.escape:
                escaped_full_packet.extend((Packet.Encoding.escape, Packet.Encoding.escaped_escape))
            elif i == Packet.Encoding.start:
                escaped_full_packet.extend((Packet.Encoding.escape, Packet.Encoding.escaped_start))
            elif i == Packet.Encoding.end:
                escaped_full_packet.extend((Packet.Encoding.escape, Packet.Encoding.escaped_end))
            else:
                escaped_full_packet.append(i)

        return [Packet.Encoding.start, *escaped_full_packet, Packet.Encoding.end]


class Collector:
    def __init__(self, callback):
        self.__callback = callback
        self.__data = []

    def add(self, data):
        for b in data:
            self.__data.append(b)
            if b == Packet.Encoding.end:
                pkt = self.__data
                self.__data = []
                if len(pkt) < 6:
                    raise PacketDecodingException(f'Very small packet {[hex(x) for x in pkt]}')
                self.__callback(Packet.from_response(pkt))



from enum import IntEnum
from functools import partial

from spherov2.packet import Packet


class BatteryStates(IntEnum):
    CHARGED = 0
    CHARGING = 1
    NOT_CHARGING = 2
    OK = 3
    LOW = 4
    CRITICAL = 5
    UNKNOWN = 255


class BatteryVoltageAndStateStates(IntEnum):
    CHARGED = 0
    CHARGING = 1
    NOT_CHARGING = 2
    OK = 3
    LOW = 4
    CRITICAL = 5
    RESERVED = 6
    UNUSED = 7


class Power:
    __encode = partial(Packet, device_id=19)

    @staticmethod
    def sleep(target_id=None):
        return Power.__encode(command_id=1, target_id=target_id)

    @staticmethod
    def get_battery_voltage(target_id=None):
        return Power.__encode(command_id=3, target_id=target_id)

    @staticmethod
    def get_battery_state(target_id=None):
        return Power.__encode(command_id=4, target_id=target_id)

    battery_state_changed_notify = (19, 6, 0xff)

    @staticmethod
    def wake(target_id=None):
        return Power.__encode(command_id=13, target_id=target_id)

    will_sleep_notify = (19, 25, 0xff)

    @staticmethod
    def enable_battery_voltage_state_change_notify(enable, target_id=None):
        return Power.__encode(command_id=27, data=[int(enable)], target_id=target_id)



from enum import IntEnum
from inspect import signature
from typing import Callable, Dict, List

from spherov2.commands.animatronic import R2LegActions
from spherov2.commands.core import IntervalOptions
from spherov2.commands.io import AudioPlaybackModes
from spherov2.commands.sensor import CollisionDetectionMethods
from spherov2.controls.enums import RawMotorModes
from spherov2.toy.bb9e import BB9E
from spherov2.toy.bolt import BOLT
from spherov2.toy.core import Toy
from spherov2.toy.mini import Mini
from spherov2.toy.r2d2 import R2D2
from spherov2.toy.r2q5 import R2Q5
from spherov2.toy.rvr import RVR


class ToyUtil:
    @staticmethod
    def sleep(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'sleep'):
            if len(signature(toy.sleep).parameters) == 0:
                toy.sleep()
            else:
                toy.sleep(IntervalOptions.NONE, 0, 0)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def ping(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'ping'):
            le = len(signature(toy.ping).parameters)
            if le == 0:
                toy.ping()
            elif le == 1:
                toy.ping(None)
            # elif l == 2: TODO
            #     toy.ping(None, ToyUtil.getPrimaryTargetId(toy))
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def roll_start(toy: Toy, heading: int, speed: int, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.roll_start(heading, speed)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def roll_stop(toy: Toy, heading: int, is_reverse: bool, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            if is_reverse:
                heading = (heading + 180) % 360
            toy.drive_control.roll_stop(heading)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def perform_leg_action(toy: Toy, leg_action: R2LegActions, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'perform_leg_action'):
            toy.perform_leg_action(leg_action)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_stabilization(toy: Toy, stabilize, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.set_stabilization(stabilize)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_raw_motor(toy: Toy, left_mode: RawMotorModes, left_speed: int, right_mode: RawMotorModes, right_speed: int,
                      not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.set_raw_motors(left_mode, left_speed, right_mode, right_speed)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def reset_heading(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.reset_heading()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def play_animation(toy: Toy, animation: IntEnum, wait: bool = False,
                       not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'play_animation'):
            toy.play_animation(animation, wait)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_head_position(toy: Toy, head_position: float, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'set_head_position'):
            toy.set_head_position(head_position)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_main_led(toy: Toy, r: int, g: int, b: int, is_user_color: bool,
                     not_supported_handler: Callable[[], None] = None):
        def _fallback():
            # TODO toys other than R2Q5
            if isinstance(toy, (R2D2, R2Q5)):
                mapping = {
                    toy.LEDs.BACK_RED: r,
                    toy.LEDs.BACK_GREEN: g,
                    toy.LEDs.BACK_BLUE: b,
                    toy.LEDs.FRONT_RED: r,
                    toy.LEDs.FRONT_GREEN: g,
                    toy.LEDs.FRONT_BLUE: b
                }
            else:
                mapping = None

            def __fallback():
                if hasattr(toy, 'set_main_led'):
                    toy.set_main_led(r, g, b)
                elif not_supported_handler:
                    not_supported_handler()

            ToyUtil.set_multiple_leds(toy, mapping, __fallback)

        ToyUtil.set_led_matrix_one_colour(toy, r, g, b, _fallback)

    @staticmethod
    def set_head_led(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, BB9E):
            ToyUtil.set_multiple_leds(toy, {BB9E.LEDs.HEAD: brightness}, not_supported_handler)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_front_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.RIGHT_HEADLIGHT_RED: r,
                RVR.LEDs.RIGHT_HEADLIGHT_GREEN: g,
                RVR.LEDs.RIGHT_HEADLIGHT_BLUE: b,
                RVR.LEDs.LEFT_HEADLIGHT_RED: r,
                RVR.LEDs.LEFT_HEADLIGHT_GREEN: g,
                RVR.LEDs.LEFT_HEADLIGHT_BLUE: b
            }
        elif isinstance(toy, (R2D2, R2Q5, BOLT)):
            mapping = {
                toy.LEDs.FRONT_RED: r,
                toy.LEDs.FRONT_GREEN: g,
                toy.LEDs.FRONT_BLUE: b
            }
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_back_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.RIGHT_BRAKELIGHT_RED: r,
                RVR.LEDs.RIGHT_BRAKELIGHT_GREEN: g,
                RVR.LEDs.RIGHT_BRAKELIGHT_BLUE: b,
                RVR.LEDs.LEFT_BRAKELIGHT_RED: r,
                RVR.LEDs.LEFT_BRAKELIGHT_GREEN: g,
                RVR.LEDs.LEFT_BRAKELIGHT_BLUE: b
            }
        elif isinstance(toy, (R2D2, R2Q5, BOLT)):
            mapping = {
                toy.LEDs.BACK_RED: r,
                toy.LEDs.BACK_GREEN: g,
                toy.LEDs.BACK_BLUE: b
            }
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_back_led_brightness(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, (BB9E, Mini)):
            mapping = {
                toy.LEDs.AIMING: brightness
            }
        elif isinstance(toy, (R2D2, R2Q5, BOLT)):
            mapping = {
                toy.LEDs.BACK_RED: 0,
                toy.LEDs.BACK_GREEN: 0,
                toy.LEDs.BACK_BLUE: brightness,
            }
        elif isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.RIGHT_BRAKELIGHT_RED: 0,
                RVR.LEDs.RIGHT_BRAKELIGHT_GREEN: 0,
                RVR.LEDs.RIGHT_BRAKELIGHT_BLUE: brightness,
                RVR.LEDs.LEFT_BRAKELIGHT_RED: 0,
                RVR.LEDs.LEFT_BRAKELIGHT_GREEN: 0,
                RVR.LEDs.LEFT_BRAKELIGHT_BLUE: brightness
            }
        else:
            mapping = None

        def _fallback():
            if hasattr(toy, 'set_back_led_brightness'):
                toy.set_back_led_brightness(brightness)
            elif not_supported_handler:
                not_supported_handler()

        ToyUtil.set_multiple_leds(toy, mapping, _fallback)

    @staticmethod
    def set_holo_projector(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, (R2D2, R2Q5)):
            mapping = {toy.LEDs.HOLO_PROJECTOR: brightness}
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_logic_display(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, (R2D2, R2Q5)):
            mapping = {toy.LEDs.LOGIC_DISPLAYS: brightness}
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_multiple_leds(toy: Toy, mapping: Dict[IntEnum, int], not_supported_handler: Callable[[], None] = None):
        if mapping and hasattr(toy, 'multi_led_control'):
            toy.multi_led_control.set_leds(mapping)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_led_matrix_one_colour(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'set_compressed_frame_player_one_color'):
            toy.set_compressed_frame_player_one_color(r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def turn_off_leds(toy: Toy):
        if hasattr(toy, 'LEDs'):
            mapping = dict((e, 0) for e in toy.LEDs)
        else:
            mapping = None

        def __fallback():
            ToyUtil.set_main_led(toy, 0, 0, 0, False)
            ToyUtil.set_back_led_brightness(toy, 0)

        ToyUtil.set_multiple_leds(toy, mapping, __fallback)
        ToyUtil.set_led_matrix_one_colour(toy, 0, 0, 0)

    @staticmethod
    def play_sound(toy: Toy, sound: IntEnum, force_play: bool, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'play_audio_file'):
            toy.play_audio_file(sound,
                                AudioPlaybackModes.PLAY_IMMEDIATELY if force_play
                                else AudioPlaybackModes.PLAY_ONLY_IF_NOT_PLAYING)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def enable_sensors(toy: Toy, sensors: List[str], not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'sensor_control'):
            toy.sensor_control.enable(*sensors)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def disable_sensors(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'sensor_control'):
            toy.sensor_control.disable_all()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def add_listeners(toy: Toy, manager):
        if hasattr(toy, 'sensor_control') and hasattr(manager, '_sensor_data_listener'):
            toy.sensor_control.add_sensor_data_listener(manager._sensor_data_listener)
        if hasattr(toy, 'add_collision_detected_notify_listener') and hasattr(manager, '_collision_detected_notify'):
            toy.add_collision_detected_notify_listener(manager._collision_detected_notify)
        if hasattr(toy, 'add_battery_state_changed_notify_listener') and \
                hasattr(manager, '_battery_state_changed_notify'):
            toy.add_battery_state_changed_notify_listener(manager._battery_state_changed_notify)
        if hasattr(toy, 'add_gyro_max_notify_listener') and hasattr(manager, '_gyro_max_notify'):
            toy.add_gyro_max_notify_listener(manager._gyro_max_notify)
        if hasattr(toy, 'add_will_sleep_notify_listener') and hasattr(manager, '_will_sleep_notify'):
            toy.add_will_sleep_notify_listener(manager._will_sleep_notify)

    @staticmethod
    def set_locator_flags(toy: Toy, flag: bool, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'set_locator_flags'):
            toy.set_locator_flags(flag)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def reset_locator(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'configure_locator'):
            toy.configure_locator(0, 0, 0, 0)
        elif hasattr(toy, 'reset_locator_x_and_y'):
            toy.reset_locator_x_and_y()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def configure_collision_detection(toy: Toy, not_supported_handler: Callable[[], None] = None):
        x, y, t = 90, 130, 1
        if hasattr(toy, 'configure_collision_detection'):
            toy.configure_collision_detection(CollisionDetectionMethods.ACCELEROMETER_BASED_DETECTION, x, y, x, y, t)
        # TODO other toys
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_power_notifications(toy: Toy, enable: bool, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'enable_charger_state_changed_notify'):
            toy.enable_charger_state_changed_notify(enable)
        elif hasattr(toy, 'enable_battery_state_changed_notify'):
            toy.enable_battery_state_changed_notify(enable)
        elif hasattr(toy, 'enable_battery_voltage_state_change_notify'):
            toy.enable_battery_voltage_state_change_notify(enable)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_robot_state_on_start(toy: Toy):
        # TODO setUserColour
        ToyUtil.set_head_position(toy, 0)
        ToyUtil.perform_leg_action(toy, R2LegActions.THREE_LEGS)
        ToyUtil.set_locator_flags(toy, False)
        ToyUtil.configure_collision_detection(toy)
        ToyUtil.set_power_notifications(toy, True)
        if hasattr(toy, 'enable_gyro_max_notify'):
            toy.enable_gyro_max_notify(True)
        if hasattr(toy, 'sensor_control'):
            toy.sensor_control.set_interval(150)
        ToyUtil.turn_off_leds(toy)
        '''if (toy instanceof HasSpheroRVRToy) {
                setColorDetection$default(toy, true, (Function0) null, 2, (Object) null);
                resetHeading$default(toy, (Function0) null, 1, (Object) null);
            }'''
        ToyUtil.reset_locator(toy)



from enum import IntEnum
from functools import partial

from spherov2.helper import to_bytes
from spherov2.packet import Packet


class CollisionDetectionMethods(IntEnum):
    NO_COLLISION_DETECTION = 0
    ACCELEROMETER_BASED_DETECTION = 1
    ACCELEROMETER_BASED_WITH_EXTRA_FILTERING = 2
    HYBRID_ACCELEROMETER_AND_CONTROL_SYSTEM_DETECTION = 3


class Sensor:
    __encode = partial(Packet, device_id=24)

    @staticmethod
    def set_sensor_streaming_mask(interval, count, sensor_masks, target_id=None):
        return Sensor.__encode(
            command_id=0,
            data=[*to_bytes(interval, 2), count, *to_bytes(sensor_masks, 4)],
            target_id=target_id
        )

    sensor_streaming_data_notify = (24, 2, 0xff)

    @staticmethod
    def set_extended_sensor_streaming_mask(sensor_masks, target_id=None):
        return Sensor.__encode(command_id=12, data=to_bytes(sensor_masks, 4), target_id=target_id)

    @staticmethod
    def enable_gyro_max_notify(enable, target_id=None):
        return Sensor.__encode(command_id=15, data=[int(enable)], target_id=target_id)

    gyro_max_notify = (24, 16, 0xff)

    @staticmethod
    def configure_collision_detection(collision_detection_method: CollisionDetectionMethods,
                                      x_threshold, y_threshold, x_speed, y_speed, dead_time, target_id=None):
        return Sensor.__encode(command_id=17,
                               data=[collision_detection_method, x_threshold, y_threshold, x_speed, y_speed, dead_time],
                               target_id=target_id)

    collision_detected_notify = (24, 18, 0xff)

    @staticmethod
    def reset_locator_x_and_y(target_id=None):
        return Sensor.__encode(command_id=19, target_id=target_id)

    @staticmethod
    def set_locator_flags(locator_flags: bool, target_id=None):
        return Sensor.__encode(command_id=23, data=[int(locator_flags)], target_id=target_id)
